/*
 * Copyright (c) 2010 Ingres Corp.
 */

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>

EXEC SQL INCLUDE SQLCA;

static void usage(char *msg)
{
	if (msg)
            fprintf(stderr, "%s\n", msg);
	fprintf( stderr, "TPC-H ACID Transaction\n\n");
	fprintf( stderr, "usage: acidtrans -b <database> [-s <n>] < <key file>\n");
	fprintf( stderr, "  -s <n>   sleep n seconds prior to commit\n\n");
    fprintf( stderr, "\te.g. acidkeys -b sf1 | acidtrans -b sf1\n\n");
	exit(0);
}

int main( int argc, char **argv)
{
	EXEC SQL BEGIN DECLARE SECTION;
		char *database=NULL;
        long long o_key;
        double tax, disc;
        long long quantity;
        double ototal, extprice;
        long long pkey, skey;
        long long l_key;
        int delta;
        double new_ototal;
        double rprice;
	EXEC SQL END DECLARE SECTION;

    char c;
	struct timeval tp;
	struct tm* ptm;
	char time_string[40];

    int delay = 0;

	extern char *optarg;
    extern int optind, optopt;

	while ((c = getopt(argc, argv, ":b:s:")) != -1)
		switch (c) {
		case 'b':
			database = optarg;
			break;
		case 's':
			delay = atoi(optarg);
			break;
		case ':':  /* -b without arguments */
			usage("option requires an argument");
			break;
		case '?':
			usage("unknown option");
		}

	if (database == NULL)
		usage("no database specified");

	if (delay < 0)
		usage("-s must be greater than or equal to 0");

	gettimeofday(&tp, NULL);
	ptm = localtime(&tp.tv_sec);
	strftime(time_string, sizeof(time_string), "%Y-%m-%d %H:%M:%S", ptm);
	printf("=========================================================\n");
	printf("TPC-H ACID Program started at: %s.%06ld\n", time_string, tp.tv_usec);
	printf("=========================================================\n");

	EXEC SQL WHENEVER SQLERROR STOP;
	EXEC SQL CONNECT TO :database;
    EXEC SQL SET AUTOCOMMIT OFF;

    while (scanf("%ld %ld %d\n", &o_key, &l_key, &delta) == 3) {
        EXEC SQL SELECT O_TOTALPRICE
		INTO :ototal
		FROM ORDERS WHERE O_ORDERKEY = :o_key;

        EXEC SQL SELECT L_QUANTITY, L_EXTENDEDPRICE, L_PARTKEY,
        L_SUPPKEY, L_TAX, L_DISCOUNT, TRUNC(L_EXTENDEDPRICE/L_QUANTITY, 2)
		INTO :quantity, :extprice, :pkey, :skey, :tax, :disc, :rprice
        FROM LINEITEM WHERE L_ORDERKEY = :o_key AND L_LINENUMBER = :l_key;

//ACID CODE BEGIN
		gettimeofday(&tp, NULL);
		ptm = localtime(&tp.tv_sec);
		strftime(time_string, sizeof(time_string), "%Y-%m-%d %H:%M:%S", ptm);

		printf("===========================================================\n");
		printf("TPC-H ACID Transaction started at: %s.%06ld\n", time_string, tp.tv_usec);
		printf("===========================================================\n");
		printf("O_TOTALPRICE from ORDERS before COMMIT: %f\n\n", ototal);

		printf("Data from LINEITEM before changes:\n");
		printf("L_QUANTITY = %ld\n", quantity);
		printf("L_EXTENDEDPRICE = %f\n", extprice);
		printf("L_PARTKEY = %ld\n", pkey);
		printf("L_SUPPKEY = %ld\n", skey);
		printf("L_TAX = %f\n", tax);
		printf("L_DISCOUNT = %f\n\n", disc);
		fflush(NULL);
//ACID CODE END

		EXEC SQL SELECT TRUNC( TRUNC(L_EXTENDEDPRICE * (1.0 - L_DISCOUNT), 2) * (1.0 + L_TAX),2 )
		INTO :new_ototal
		FROM LINEITEM WHERE L_ORDERKEY = :o_key AND l_linenumber = :l_key;

		printf("Select into new_ototal complete\n"); fflush(stdout);
		// Update ORDERS a first time to remove old LINEITEM from total price
		EXEC SQL UPDATE ORDERS
		SET O_TOTALPRICE = O_TOTALPRICE - decimal(:new_ototal,12,2)
		WHERE O_ORDERKEY = :o_key;
		printf("Update orders complete\n"); fflush(stdout);

		// Update LINEITEM to increase quantity
		EXEC SQL UPDATE LINEITEM
		SET L_EXTENDEDPRICE = L_EXTENDEDPRICE + TRUNC(TRUNC(L_EXTENDEDPRICE/L_QUANTITY, 2) * :delta, 2 ),
		L_QUANTITY = L_QUANTITY + :delta
		WHERE L_ORDERKEY = :o_key AND L_LINENUMBER = :l_key;
		printf("Update lineitem complete\n"); fflush(stdout);

		EXEC SQL SELECT TRUNC( TRUNC(L_EXTENDEDPRICE * (1.0 - L_DISCOUNT), 2) * (1.0 + L_TAX),2 )
		INTO :new_ototal
		FROM LINEITEM WHERE L_ORDERKEY = :o_key AND l_linenumber = :l_key;
		printf("Select new total into new_ototal complete\n"); fflush(stdout);

		// Update ORDERS a second time to add updated LINEITEM to total price
		EXEC SQL UPDATE ORDERS
		SET O_TOTALPRICE = O_TOTALPRICE + decimal(:new_ototal,12,2)
		WHERE O_ORDERKEY = :o_key;
		printf("2nd update orders complete\n"); fflush(stdout);

		// Insert into HISTORY records of transaction
        EXEC SQL INSERT INTO HISTORY
        VALUES( :pkey, :skey, :o_key, :l_key, :delta, timestamp(date('now')));
	printf("Insert into history complete\n"); fflush(stdout);

//ACID CODE BEGIN
        if (delay) {
    		gettimeofday(&tp, NULL);
    		ptm = localtime(&tp.tv_sec);
    		strftime(time_string, sizeof(time_string), "%Y-%m-%d %H:%M:%S", ptm);

    		printf(">> TPC-H ACID Transaction. Start of SLEEP: %s.%06ld\n\n", time_string, tp.tv_usec);
    		fflush(NULL);

        	sleep(delay);

    		gettimeofday(&tp, NULL);
    		ptm = localtime(&tp.tv_sec);
    		strftime(time_string, sizeof(time_string), "%Y-%m-%d %H:%M:%S", ptm);

    		printf(">> TPC-H ACID Transaction. End of SLEEP: %s.%06ld\n\n", time_string, tp.tv_usec);
    		fflush(NULL);
        }
//ACID CODE END

        EXEC SQL COMMIT WORK;

//ACID CODE BEGIN
		gettimeofday(&tp, NULL);
		ptm = localtime(&tp.tv_sec);
		strftime(time_string, sizeof(time_string), "%Y-%m-%d %H:%M:%S", ptm);

        // write txn details into a success record
		printf("success_record==> %ld\t%ld\t%ld\t%ld\t%d\t%s.%06ld\n", pkey, skey, o_key, l_key, delta, time_string, tp.tv_usec);
		fflush(stdout);

        EXEC SQL SELECT O_TOTALPRICE
		INTO :ototal
		FROM ORDERS WHERE O_ORDERKEY = :o_key;

        EXEC SQL SELECT L_QUANTITY, L_EXTENDEDPRICE, L_PARTKEY, L_SUPPKEY, L_TAX, L_DISCOUNT
		INTO :quantity, :extprice, :pkey, :skey, :tax, :disc
        FROM LINEITEM WHERE L_ORDERKEY = :o_key AND L_LINENUMBER = :l_key;

		printf("O_TOTALPRICE from ORDERS after COMMIT: %f\n\n", ototal);

		printf("Data from LINEITEM after COMMIT:\n");
		printf("L_QUANTITY = %ld\n", quantity);
		printf("L_EXTENDEDPRICE = %f\n", extprice);
		printf("L_PARTKEY = %ld\n", pkey);
		printf("L_SUPPKEY = %ld\n", skey);
		printf("L_TAX = %f\n", tax);
		printf("L_DISCOUNT = %f\n\n", disc);

		gettimeofday(&tp, NULL);
		ptm = localtime(&tp.tv_sec);
		strftime(time_string, sizeof(time_string), "%Y-%m-%d %H:%M:%S", ptm);

		printf(">> TPC-H ACID Transaction Ended at: %s.%06ld\n\n", time_string, tp.tv_usec);
//ACID CODE END

        printf("%15s %10s %6s %6s %15s %15s\n",
               "rprice", "quantity", "tax", "disc", "extprice", "ototal");
        printf("%15.2f %10ld %6.2f %6.2f %15.2f %15.2f\n",
                rprice, quantity, tax, disc, extprice, ototal);
        fflush(stdout);
    }
    EXEC SQL DISCONNECT;

	gettimeofday(&tp, NULL);
	ptm = localtime(&tp.tv_sec);
	strftime(time_string, sizeof(time_string), "%Y-%m-%d %H:%M:%S", ptm);
	printf("=======================================================\n");
	printf("TPC-H ACID Program ended at: %s.%06ld\n", time_string, tp.tv_usec);
	printf("=======================================================\n");

    return(0);
}
